class star:
    # Base class for a single action in a formation. Contains nodes this star point to.
    def __init__(self):
        self._nodes = []

    def link(self, star2):
        # Chain stars. Execution star.action() -> star2.action()
        self._nodes.append(star2)
        return self

    def unlink(self, star2):
        # Chain stars. Execution star.action() -> star2.action()
        self._nodes.remove(star2)
        return self

    def action(self, s):
        return s



from starformation.star import star


class CStar(star):
    # Wrap custom function
    def __init__(self, func):
        super().__init__()
        self.func = func

    def action(self, *args, **kwargs):
        return self.func(*args, **kwargs)




from .star import star
import pathos.multiprocessing as mp

class Formation:
    # Contains multiple stars; Chains them and run them.
    def __init__(self, star=None):
        self.head = star
        self.end = star

    def add(self, s: star):
        # Add first star in formation or chain stars to the last added star.
        if self.head is None:
            self.head = s
            self.end = s
        else:
            self.end.link(s)
            self.end = s

    def mp_run(self, input, cores):
        # run all process on multi-processes and save output
        def run_star(star, input):
            # Run all the star and save output of last stars
            output = star.action(input)
            return output

        stars = [(self.head, input)]
        while len(stars):
            with mp.Pool(cores) as p:
                outputs = p.starmap(run_star, [(i, input) for i, input in stars])
                next_stars = []
                for outputNum, (star, _) in enumerate(stars):
                    next_stars += zip(star._nodes, [outputs[outputNum]] * len(star._nodes))
                stars = next_stars
        return outputs

    def run(self, input = None, multicpu:int = 0):
        if multicpu > 1:
            return self.mp_run(input, multicpu)
        # run all star actions and return output
        outputs=[]

        def run_star(star, input):
            # Run all the star and save output of last stars
            output = star.action(input)
            for i in star._nodes:
                run_star(i, output)
            if len(star._nodes) == 0:
                outputs.append(output)

        run_star(self.head, input)
        return outputs



import setuptools
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="starformation",
    version="0.0.3",
    author="Tommy Dong",
    author_email="tommydong1998@gmail.com",
    description="Star Formation is a Python library to quickly develop clean modular models/pipelines.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TommyDong1998/StarFormation",
    packages=['starformation','starformation.classic'],
    install_requires=['backtrader','pathos']
)